
// Demo program that demonstrates how the neck servos and IR head can be used
//
// This program should cause the head to follow your hand when it's up close.
// This program also continually prints out the amount of IR light detected by 
// the sensors on the IR head in the order

//           Left, Right, Top, Bottom

//------------------------------------------------------------------------------
#include <Servo.h>

//------------------------------------------------------------------------------
const int SPEAKER_PIN = 13;
const int HEAD_LEFT_IR_SENSOR_PIN = A4;
const int HEAD_RIGHT_IR_SENSOR_PIN = A5;
const int HEAD_TOP_IR_SENSOR_PIN = A6;
const int HEAD_BOTTOM_IR_SENSOR_PIN = A7;
const int HEAD_IR_LEDS_PIN = 2;

const int NECK_PAN_PIN = 5;
const int NECK_TILT_PIN = 6;

const int NATURAL_LEFT_RIGHT_IR_DIFF = 50;
const int NATURAL_TOP_BOTTOM_IR_DIFF = 0;
const int NECK_PAN_CENTRE_ANGLE = 85;
const int NECK_PAN_MIN_ANGLE = 0;        // This faces right
const int NECK_PAN_MAX_ANGLE = 180;      // This faces left
const int NECK_TILT_CENTRE_ANGLE = 90;
const int NECK_TILT_MIN_ANGLE = 50;       // This faces down
const int NECK_TILT_MAX_ANGLE = 180;      // This faces up

const int NECK_RESET_TIME_MS = 2000;    // If we don't see anything for this amount of time
                                        // the neck is reset

const int MIN_IR = 250; 
const float MIN_ABS_DIFF_NEEDED_FOR_TRACKING = 10.0f;

const float MAX_ABS_NECK_ANGLE_CHANGE = 3.0f;    // These to constants control how 
const float DIFF_TO_ANGLE_SCALE = 0.01f;         // fast the neck will move



struct IREyeReading
{
    int mLeftIR;
    int mRightIR;
    int mTopIR;
    int mBottomIR;
    int mAverageIRFromObject;
}; 

Servo gNeckPanServo;
Servo gNeckTiltServo;
unsigned long gLastIRTime = 0;
float gNeckPanServoAngle = NECK_PAN_CENTRE_ANGLE;
float gNeckTiltServoAngle = NECK_TILT_CENTRE_ANGLE;

//------------------------------------------------------------------------------
void ReadFromIREye( IREyeReading* pReadingOut ); 

//------------------------------------------------------------------------------
void setup()
{
    // Initialise servos and configure pins
    gNeckPanServo.attach( NECK_PAN_PIN );
    gNeckTiltServo.attach( NECK_TILT_PIN );
    
    pinMode( SPEAKER_PIN, OUTPUT );
    pinMode( HEAD_IR_LEDS_PIN, OUTPUT );
    
    Serial.begin( 9600 );
}

//------------------------------------------------------------------------------
void loop()
{
    unsigned long curTime = millis();
    
    // Track objects using the IR eye on the head
    IREyeReading irEyeReading;
    ReadFromIREye( &irEyeReading );
    
    if ( irEyeReading.mAverageIRFromObject >= MIN_IR )
    {
        // Test to see if we should go left or right
        float leftRightDiff = (irEyeReading.mLeftIR - irEyeReading.mRightIR) - NATURAL_LEFT_RIGHT_IR_DIFF;
        if ( abs( leftRightDiff ) > MIN_ABS_DIFF_NEEDED_FOR_TRACKING )
        {
            float angleChange = leftRightDiff * DIFF_TO_ANGLE_SCALE;
            angleChange = constrain( angleChange, -MAX_ABS_NECK_ANGLE_CHANGE, MAX_ABS_NECK_ANGLE_CHANGE );

            gNeckPanServoAngle += angleChange;
        }

        
        // Test to see if we should go up or down
        float topBottomDiff = (irEyeReading.mTopIR - irEyeReading.mBottomIR) - NATURAL_TOP_BOTTOM_IR_DIFF;
        if ( abs( topBottomDiff ) > MIN_ABS_DIFF_NEEDED_FOR_TRACKING )
        {
            float angleChange = topBottomDiff * DIFF_TO_ANGLE_SCALE;
            angleChange = constrain( angleChange, -MAX_ABS_NECK_ANGLE_CHANGE, MAX_ABS_NECK_ANGLE_CHANGE );

            gNeckTiltServoAngle += angleChange;
        }
        
        // record the last time we saw something with the IR head
        gLastIRTime = curTime;
    }
    else
    {
        // Reset head if we don't seen anything for a while
        if ( curTime - gLastIRTime >= NECK_RESET_TIME_MS )
        {
            gNeckPanServoAngle = NECK_PAN_CENTRE_ANGLE;
            gNeckTiltServoAngle = NECK_TILT_CENTRE_ANGLE;
        }
    }

    gNeckPanServoAngle = constrain( gNeckPanServoAngle, NECK_PAN_MIN_ANGLE, NECK_PAN_MAX_ANGLE );
    gNeckPanServo.write( gNeckPanServoAngle );
    
    gNeckTiltServoAngle = constrain( gNeckTiltServoAngle, NECK_TILT_MIN_ANGLE, NECK_TILT_MAX_ANGLE );
    gNeckTiltServo.write( gNeckTiltServoAngle );
    
    Serial.print( irEyeReading.mLeftIR );
    Serial.print( " " );
    Serial.print( irEyeReading.mRightIR );
    Serial.print( " " );
    Serial.print( irEyeReading.mTopIR );
    Serial.print( " " );
    Serial.println( irEyeReading.mBottomIR );
}

//------------------------------------------------------------------------------
void ReadFromIREye( IREyeReading* pReadingOut )
{
    static int noiseCounter = 0;
    
    if ( NULL != pReadingOut )
    {
        digitalWrite( HEAD_IR_LEDS_PIN, HIGH ); // Turn on IR LEDs to read TOTAL IR LIGHT (ambient + reflected)
        delay(2);                               // Allow time for phototransistors to respond. (may not be needed)
        pReadingOut->mLeftIR = analogRead( HEAD_LEFT_IR_SENSOR_PIN );
        pReadingOut->mRightIR = analogRead( HEAD_RIGHT_IR_SENSOR_PIN );
        pReadingOut->mTopIR = analogRead( HEAD_TOP_IR_SENSOR_PIN );
        pReadingOut->mBottomIR = analogRead( HEAD_BOTTOM_IR_SENSOR_PIN );

        digitalWrite( HEAD_IR_LEDS_PIN, LOW );  // turn off IR LEDs to read AMBIENT IR LIGHT (IR from indoor lighting and sunlight)
        delay(2);                               // Allow time for phototransistors to respond. (may not be needed)
        pReadingOut->mLeftIR -= analogRead( HEAD_LEFT_IR_SENSOR_PIN );
        pReadingOut->mRightIR -= analogRead( HEAD_RIGHT_IR_SENSOR_PIN );
        pReadingOut->mTopIR -= analogRead( HEAD_TOP_IR_SENSOR_PIN );
        pReadingOut->mBottomIR -= analogRead( HEAD_BOTTOM_IR_SENSOR_PIN );

        pReadingOut->mAverageIRFromObject = (pReadingOut->mLeftIR + pReadingOut->mRightIR
            + pReadingOut->mTopIR + pReadingOut->mBottomIR) / 4;

        noiseCounter++; // count program loops
        if ( noiseCounter%8 == 0 )
        {
            if ( pReadingOut->mAverageIRFromObject > MIN_IR )
            {
                tone( SPEAKER_PIN, pReadingOut->mAverageIRFromObject*5+100,5); // produce sound every eighth loop - high pitch = close distance
            }
        }
    }
} 
