
// Demo program for the Adventure bot
//
// This is an alternative to the default example program from Dagu

//------------------------------------------------------------------------------
#include <Servo.h>
#include "pitches.h"

//------------------------------------------------------------------------------
const int SPEAKER_PIN = 13;
const int HEAD_LEFT_IR_SENSOR_PIN = A4;
const int HEAD_RIGHT_IR_SENSOR_PIN = A5;
const int HEAD_TOP_IR_SENSOR_PIN = A6;
const int HEAD_BOTTOM_IR_SENSOR_PIN = A7;
const int HEAD_IR_LEDS_PIN = 2;

const int NECK_PAN_PIN = 5;
const int NECK_TILT_PIN = 6;

const int FRONT_LEFT_LED_PIN = 8;
const int FRONT_RIGHT_LED_PIN = 7;
const int REAR_LEFT_LED_PIN = 12;
const int REAR_RIGHT_LED_PIN = 4;

const int FRONT_LEFT_SENSE_PIN = A2;
const int FRONT_RIGHT_SENSE_PIN = A3;
const int REAR_LEFT_SENSE_PIN = A1;
const int REAR_RIGHT_SENSE_PIN = A0;

const int LEFT_MOTOR_PIN = 9;
const int RIGHT_MOTOR_PIN = 10;
const int MIN_MOTOR_SPEED = -100;
const int MAX_MOTOR_SPEED = 100;

const int LEFT_MOTOR_FULL_REVERSE_US = 2000;
const int LEFT_MOTOR_FULL_STOP_US = 1500;
const int LEFT_MOTOR_FULL_FORWARD_US = 1000;
const int RIGHT_MOTOR_FULL_REVERSE_US = 1000;
const int RIGHT_MOTOR_FULL_STOP_US = 1500;
const int RIGHT_MOTOR_FULL_FORWARD_US = 2000;

const int OBSTACLE_DETECTED_IR_LEVEL = 100;

const int NATURAL_LEFT_RIGHT_IR_DIFF = 50;
const int NATURAL_TOP_BOTTOM_IR_DIFF = 0;
const int NECK_PAN_CENTRE_ANGLE = 85;
const int NECK_PAN_MIN_ANGLE = 0;        // This faces right
const int NECK_PAN_MAX_ANGLE = 180;      // This faces left
const int NECK_TILT_CENTRE_ANGLE = 90;
const int NECK_TILT_MIN_ANGLE = 50;       // This faces down
const int NECK_TILT_MAX_ANGLE = 180;      // This faces up

const int NECK_RESET_TIME_MS = 2000;    // If we don't see anything for this amount of time
                                        // the neck is reset

const int MIN_IR = 200; 
const float MIN_ABS_DIFF_NEEDED_FOR_TRACKING = 10.0f;

const float MAX_ABS_NECK_ANGLE_CHANGE = 3.0f;    // These to constants control how 
const float DIFF_TO_ANGLE_SCALE = 0.01f;         // fast the neck will move

const float MIN_ABS_TURN_CATCHUP_ANGLE = 20.0f;
const int BACKUP_IR = 900; 
const boolean ENABLE_BACKUP = false;

struct IREyeReading
{
    int mLeftIR;
    int mRightIR;
    int mTopIR;
    int mBottomIR;
    int mAverageIRFromObject;
}; 

Servo gNeckPanServo;
Servo gNeckTiltServo;
unsigned long gLastIRTime = 0;
float gNeckPanServoAngle = NECK_PAN_CENTRE_ANGLE;
float gNeckTiltServoAngle = NECK_TILT_CENTRE_ANGLE;

Servo gLeftMotor;
Servo gRightMotor;

//------------------------------------------------------------------------------
void UpdateCornerSensors(); 
void ReadFromIREye( IREyeReading* pReadingOut ); 
int convertLeftMotorSpeedToUS( int motorSpeed );
int convertRightMotorSpeedToUS( int motorSpeed );

//------------------------------------------------------------------------------
void setup()
{
    // Initialise servos and configure pins
    gNeckPanServo.attach( NECK_PAN_PIN );
    gNeckTiltServo.attach( NECK_TILT_PIN );
    gLeftMotor.attach( LEFT_MOTOR_PIN );
    gRightMotor.attach( RIGHT_MOTOR_PIN );
    
    pinMode( SPEAKER_PIN, OUTPUT );
    pinMode( HEAD_IR_LEDS_PIN, OUTPUT );
    
    // Play tune on powerup / reset
    int melody[] = { NOTE_C4, NOTE_G3, NOTE_G3, NOTE_A3, NOTE_G3, 0, NOTE_B3, NOTE_C4 };
    int noteDurations[] = { 4, 8, 8, 4, 4, 4, 4, 4 };
    for ( byte note = 0; note < 8; note++ )
    {
        int noteDuration = 1000/noteDurations[ note ];
        tone( SPEAKER_PIN, melody[note], noteDuration );
        int pauseBetweenNotes = noteDuration * 1.30;
        delay( pauseBetweenNotes );
    }
}

//------------------------------------------------------------------------------
void loop()
{
    unsigned long curTime = millis();
    
    UpdateCornerSensors();
    
    // Track objects using the IR eye on the head
    IREyeReading irEyeReading;
    ReadFromIREye( &irEyeReading );
    
    if ( irEyeReading.mAverageIRFromObject >= MIN_IR )
    {
        // Test to see if we should go left or right
        float leftRightDiff = (irEyeReading.mLeftIR - irEyeReading.mRightIR) - NATURAL_LEFT_RIGHT_IR_DIFF;
        if ( abs( leftRightDiff ) > MIN_ABS_DIFF_NEEDED_FOR_TRACKING )
        {
            float angleChange = leftRightDiff * DIFF_TO_ANGLE_SCALE;
            angleChange = constrain( angleChange, -MAX_ABS_NECK_ANGLE_CHANGE, MAX_ABS_NECK_ANGLE_CHANGE );

            gNeckPanServoAngle += angleChange;
        }

        
        // Test to see if we should go up or down
        float topBottomDiff = (irEyeReading.mTopIR - irEyeReading.mBottomIR) - NATURAL_TOP_BOTTOM_IR_DIFF;
        if ( abs( topBottomDiff ) > MIN_ABS_DIFF_NEEDED_FOR_TRACKING )
        {
            float angleChange = topBottomDiff * DIFF_TO_ANGLE_SCALE;
            angleChange = constrain( angleChange, -MAX_ABS_NECK_ANGLE_CHANGE, MAX_ABS_NECK_ANGLE_CHANGE );

            gNeckTiltServoAngle += angleChange;
        }
        
        // record the last time we saw something with the IR head
        gLastIRTime = curTime;
    }
    else
    {
        // Reset head if we don't seen anything for a while
        if ( curTime - gLastIRTime >= NECK_RESET_TIME_MS )
        {
            gNeckPanServoAngle = NECK_PAN_CENTRE_ANGLE;
            gNeckTiltServoAngle = NECK_TILT_CENTRE_ANGLE;
        }
    }

    gNeckPanServoAngle = constrain( gNeckPanServoAngle, NECK_PAN_MIN_ANGLE, NECK_PAN_MAX_ANGLE );
    gNeckPanServo.write( gNeckPanServoAngle );
    
    gNeckTiltServoAngle = constrain( gNeckTiltServoAngle, NECK_TILT_MIN_ANGLE, NECK_TILT_MAX_ANGLE );
    gNeckTiltServo.write( gNeckTiltServoAngle );
    
    // Turn the body to match the head
    float neckAngleFromCentre = gNeckPanServoAngle - NECK_PAN_CENTRE_ANGLE;
    if ( neckAngleFromCentre > MIN_ABS_TURN_CATCHUP_ANGLE )
    {
        // Turn left to catch up
        gLeftMotor.writeMicroseconds( convertLeftMotorSpeedToUS( -30 ) );
        gRightMotor.writeMicroseconds( convertRightMotorSpeedToUS( 30 ) );
    }
    else if ( neckAngleFromCentre < -MIN_ABS_TURN_CATCHUP_ANGLE )
    {
        // Turn right to catch up
        gLeftMotor.writeMicroseconds( convertLeftMotorSpeedToUS( 30 ) );
        gRightMotor.writeMicroseconds( convertRightMotorSpeedToUS( -30 ) );
    }
    else
    {
        // Either backup or halt the motors
        if ( irEyeReading.mAverageIRFromObject >= BACKUP_IR && ENABLE_BACKUP )
        {
            gLeftMotor.writeMicroseconds( convertLeftMotorSpeedToUS( -20 ) );
            gRightMotor.writeMicroseconds( convertRightMotorSpeedToUS( -20 ) );
        }
        else
        {
            gLeftMotor.writeMicroseconds( convertLeftMotorSpeedToUS( 0 ) );
            gRightMotor.writeMicroseconds( convertRightMotorSpeedToUS( 0 ) );
        }
    }
}

//------------------------------------------------------------------------------
void UpdateCornerSensors()
{
    // Turn on the IR  LEDs
    digitalWrite( FRONT_LEFT_LED_PIN, 1 );
    digitalWrite( FRONT_RIGHT_LED_PIN, 1 );
    digitalWrite( REAR_LEFT_LED_PIN, 1 );
    digitalWrite( REAR_RIGHT_LED_PIN, 1 );
    
    delay( 2 );    // Give sensors time to respond
    
    // Read total IR values
    int frontLeftIR = analogRead( FRONT_LEFT_SENSE_PIN );
    int frontRightIR = analogRead( FRONT_RIGHT_SENSE_PIN );
    int rearLeftIR = analogRead( REAR_LEFT_SENSE_PIN );
    int rearRightIR = analogRead( REAR_RIGHT_SENSE_PIN );
  
    // Turn off the IR LEDs
    digitalWrite( FRONT_LEFT_LED_PIN, 0 );
    digitalWrite( FRONT_RIGHT_LED_PIN, 0 );
    digitalWrite( REAR_LEFT_LED_PIN, 0 );
    digitalWrite( REAR_RIGHT_LED_PIN, 0 );

    delay( 2 );    // Give sensors time to respond

    // Subtract ambient IR from total IR to give reflected IR values
    frontLeftIR -= analogRead( FRONT_LEFT_SENSE_PIN );
    frontRightIR -= analogRead( FRONT_RIGHT_SENSE_PIN );
    rearLeftIR -= analogRead( REAR_LEFT_SENSE_PIN );
    rearRightIR -= analogRead( REAR_RIGHT_SENSE_PIN );
    
    // If an obstacle is close, then the IR value will be above a certain threshold
    digitalWrite( FRONT_LEFT_LED_PIN, frontLeftIR > OBSTACLE_DETECTED_IR_LEVEL );
    digitalWrite( FRONT_RIGHT_LED_PIN, frontRightIR > OBSTACLE_DETECTED_IR_LEVEL );
    digitalWrite( REAR_LEFT_LED_PIN, rearLeftIR > OBSTACLE_DETECTED_IR_LEVEL );
    digitalWrite( REAR_RIGHT_LED_PIN, rearRightIR > OBSTACLE_DETECTED_IR_LEVEL );
}

//------------------------------------------------------------------------------
void ReadFromIREye( IREyeReading* pReadingOut )
{
    static int noiseCounter = 0;
    
    if ( NULL != pReadingOut )
    {
        digitalWrite( HEAD_IR_LEDS_PIN, HIGH ); // Turn on IR LEDs to read TOTAL IR LIGHT (ambient + reflected)
        delay(2);                               // Allow time for phototransistors to respond. (may not be needed)
        pReadingOut->mLeftIR = analogRead( HEAD_LEFT_IR_SENSOR_PIN );
        pReadingOut->mRightIR = analogRead( HEAD_RIGHT_IR_SENSOR_PIN );
        pReadingOut->mTopIR = analogRead( HEAD_TOP_IR_SENSOR_PIN );
        pReadingOut->mBottomIR = analogRead( HEAD_BOTTOM_IR_SENSOR_PIN );

        digitalWrite( HEAD_IR_LEDS_PIN, LOW );  // turn off IR LEDs to read AMBIENT IR LIGHT (IR from indoor lighting and sunlight)
        delay(2);                               // Allow time for phototransistors to respond. (may not be needed)
        pReadingOut->mLeftIR -= analogRead( HEAD_LEFT_IR_SENSOR_PIN );
        pReadingOut->mRightIR -= analogRead( HEAD_RIGHT_IR_SENSOR_PIN );
        pReadingOut->mTopIR -= analogRead( HEAD_TOP_IR_SENSOR_PIN );
        pReadingOut->mBottomIR -= analogRead( HEAD_BOTTOM_IR_SENSOR_PIN );

        pReadingOut->mAverageIRFromObject = (pReadingOut->mLeftIR + pReadingOut->mRightIR
            + pReadingOut->mTopIR + pReadingOut->mBottomIR) / 4;

        noiseCounter++; // count program loops
        if ( noiseCounter%8 == 0 )
        {
            if ( pReadingOut->mAverageIRFromObject > MIN_IR )
            {
                tone( SPEAKER_PIN, pReadingOut->mAverageIRFromObject*5+100,5); // produce sound every eighth loop - high pitch = close distance
            }
        }
    }
} 

//------------------------------------------------------------------------------
int convertLeftMotorSpeedToUS( int motorSpeed )
{
    int pulseWidthUS = LEFT_MOTOR_FULL_STOP_US;    // Default result is full stop
    
    motorSpeed = constrain( motorSpeed, MIN_MOTOR_SPEED, MAX_MOTOR_SPEED );
    
    if ( motorSpeed < 0 )
    {
        float proportion = (float)motorSpeed/(float)MIN_MOTOR_SPEED;
        pulseWidthUS = LEFT_MOTOR_FULL_STOP_US 
            + proportion*(LEFT_MOTOR_FULL_REVERSE_US - LEFT_MOTOR_FULL_STOP_US);
    }
    else if ( motorSpeed > 0 )
    {
        float proportion = (float)motorSpeed/(float)MAX_MOTOR_SPEED;
        pulseWidthUS = LEFT_MOTOR_FULL_STOP_US 
            + proportion*(LEFT_MOTOR_FULL_FORWARD_US - LEFT_MOTOR_FULL_STOP_US);
    }
    
    return pulseWidthUS;
}

//------------------------------------------------------------------------------
int convertRightMotorSpeedToUS( int motorSpeed )
{
    int pulseWidthUS = RIGHT_MOTOR_FULL_STOP_US;    // Default result is full stop
    
    motorSpeed = constrain( motorSpeed, MIN_MOTOR_SPEED, MAX_MOTOR_SPEED );
    
    if ( motorSpeed < 0 )
    {
        float proportion = (float)motorSpeed/(float)MIN_MOTOR_SPEED;
        pulseWidthUS = RIGHT_MOTOR_FULL_STOP_US 
            + proportion*(RIGHT_MOTOR_FULL_REVERSE_US - RIGHT_MOTOR_FULL_STOP_US);
    }
    else if ( motorSpeed > 0 )
    {
        float proportion = (float)motorSpeed/(float)MAX_MOTOR_SPEED;
        pulseWidthUS = RIGHT_MOTOR_FULL_STOP_US 
            + proportion*(RIGHT_MOTOR_FULL_FORWARD_US - RIGHT_MOTOR_FULL_STOP_US);
    }
    
    return pulseWidthUS;
}
